__title__ = "MindsDB Aerospike handler"
__package_name__ = "mindsdb_aerospike_handler"
__version__ = "0.0.1"
__description__ = "MindsDB handler for Aerospike"
__author__ = "Biswadip Paul"
__github__ = "https://github.com/mindsdb/mindsdb"
__pypi__ = "https://pypi.org/project/mindsdb/"
__license__ = "GPL-3.0"
__copyright__ = "Copyright 2023 - mindsdb"
